#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to create a drawing look 

# (c) Dinasset 2020
#
#   History:
#
#   v1.0: 2020-10-31: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gimpfu import *
import sys, os
import time
import datetime

# define here some constant value, when required
FixWH=4000  # limit, user can reduce bt choosing evidence of strokes

# predefined colour names
NEUTRAL_GRAY=(170,170,170,255)
RED_GRAY=(200,170,170,255)
ORANGE_GRAY=(200,185,170,255)
YELLOW_GRAY=(200,200,170,255)
CHARTREUSE_GRAY=(185,200,170,255)
GREEN_GRAY=(170,200,170,255)
SPRING_GRAY=(170,200,185,255)
CYAN_GRAY=(170,200,200,255)
AZURE_GRAY=(170,185,200,255)
BLUE_GRAY=(170,170,200,255)
VIOLET_GRAY=(185,170,200,255)
MAGENTA_GRAY=(200,170,200,255)
ROSE_GRAY=(200,170,185,255)

# ================ *
# SERVICE ROUTINES *
# ---------------- *
#----------------------------- 
# include, then use: erro_msg 
#----------------------------- 
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
#
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
# ======================================================
# ANCILLARY MODULE : stretch R, G, B colours  
# ------------------------------------------------------
def StretchRGBcolours (layer, target) :

    img = layer.image
    targetB = target  
    targetW = target  

    # ============================
    # LEVELS STRETCH FOR COLOURS
    # ============================
    # RED
     
    fromV = 0.00   
    toV = 0.01   
    percent = 0
    while (percent < targetB) :  
        (mean,s,m,p,c,percentile) = pdb.gimp_drawable_histogram (layer, HISTOGRAM_RED, fromV, toV)
        percent = percentile * 100  
        toV += 0.01
        if toV > 1: toV = 1.0
    fromMinRED = toV - 0.01    
    
    fromV = 0.99    
    toV = 1.00    
    percent = 0
    while (percent < targetW) :  
        (mean,s,m,p,c,percentile) = pdb.gimp_drawable_histogram (layer, HISTOGRAM_RED, fromV, toV)
        percent = percentile * 100  
        fromV -= 0.01    
    toMaxRED = fromV + 0.01   
    
    # GREEN
     
    fromV = 0.00    
    toV = 0.01   
    percent = 0 
    while (percent < targetB) :  
        (mean,s,m,p,c,percentile) = pdb.gimp_drawable_histogram (layer, HISTOGRAM_GREEN, fromV, toV)
        percent = percentile * 100  
        toV += 0.01
        if toV > 1: toV = 1.0
    fromMinGREEN = toV - 0.01  
    
    fromV = 0.99 
    toV = 1.00   
    percent = 0
    while (percent < targetW) :  
        (mean,s,m,p,c,percentile) = pdb.gimp_drawable_histogram (layer, HISTOGRAM_GREEN, fromV, toV)
        percent = percentile * 100  
        fromV -= 0.01   
    toMaxGREEN = fromV + 0.01  
    
    # BLUE 
     
    fromV = 0.00  
    toV = 0.01    
    percent = 0
    while (percent < targetB) :  
        (mean,s,m,p,c,percentile) = pdb.gimp_drawable_histogram (layer, HISTOGRAM_BLUE, fromV, toV)
        percent = percentile * 100  
        toV += 0.01
        if toV > 1: toV = 1.0
    fromMinBLUE = toV - 0.01   
    
    fromV = 0.99  
    toV = 1.00   
    percent = 0
    while (percent < targetW) :  
        (mean,s,m,p,c,percentile) = pdb.gimp_drawable_histogram (layer, HISTOGRAM_BLUE, fromV, toV)
        percent = percentile * 100  
        fromV -= 0.01  
    toMaxBLUE = fromV + 0.01   
    
    # APPLY MODIFICATIONS TO LEVELS
    # but removing darks, because all drawing lines are black
    pdb.gimp_drawable_levels (layer, HISTOGRAM_RED,
                              fromMinRED, toMaxRED, False, 1.0, 0.20, 1.0, False)
    pdb.gimp_drawable_levels (layer, HISTOGRAM_GREEN,
                              fromMinGREEN, toMaxGREEN, False, 1.0, 0.20, 1.0, False)     
    pdb.gimp_drawable_levels (layer, HISTOGRAM_BLUE,
                              fromMinBLUE, toMaxBLUE, False, 1.0, 0.20, 1.0, False)     
         
    return(layer)
#
# ======================================================
# ANCILLARY MODULE : convert RGB to HSV 
# ------------------------------------------------------

def rgb2hsv(rgb):
    (r,g,b,a) = rgb
    r, g, b = r/255.0, g/255.0, b/255.0
    mx = max(r, g, b)
    mn = min(r, g, b)
    df = mx-mn
    if mx == mn:
        h = 0
    elif mx == r:
        h = (60 * ((g-b)/df) + 360) % 360
    elif mx == g:
        h = (60 * ((b-r)/df) + 120) % 360
    elif mx == b:
        h = (60 * ((r-g)/df) + 240) % 360
    if mx == 0:
        s = 0
    else:
        s = (df/mx)*100
    v = mx*100
    #info_msg ("hue="+str(h))
    #info_msg ("sat="+str(s))
    #info_msg ("val="+str(v))
    return h, s, v

#
# ======================================================
# ANCILLARY MODULE : create paper for the BG 
# ------------------------------------------------------
def CreatePaper (width,height,hue,sat,val):
    
    WH = width+height
    PaperImg = pdb.gimp_image_new (width, height, RGB_IMAGE)
    PaperDrw = pdb.gimp_layer_new (PaperImg, width, height, RGB_IMAGE,
                                   "Generated Paper", 100.0, LAYER_MODE_NORMAL)
    PaperImg.add_layer (PaperDrw,0)
    
    pdb.plug_in_plasma (PaperImg, PaperDrw, 26, 2.0)
    pdb.gimp_drawable_curves_spline (PaperDrw, HISTOGRAM_VALUE, 4,
                                    (0.00,0.50, 1.00,0.75) 
                                    )
    pdb.plug_in_rgb_noise (PaperImg, PaperDrw, True, False,
                           0.2, 0.2, 0.2, 0.2)
    pdb.plug_in_gauss (PaperImg, PaperDrw, 5.0, 5.0, 0)
    pdb.gimp_drawable_desaturate (PaperDrw, DESATURATE_LUMA)
    pdb.gimp_drawable_colorize_hsl (PaperDrw, hue*1.0, sat*1.0, val*1.0-100.0)
    
    #pdb.gimp_display_new (PaperImg)
    
    return (PaperDrw)
#
# ================ *
#   M  A  I  N     *
# ---------------- *
   
def DrawStyleE (inImage, inBG, outimgWH,
                predefcol, inColor, inEvidence,
                inOpacity):
    
    pdb.gimp_context_push()
    
    width = inBG.width
    height = inBG.height
    OrigWH = width+height
    source_name = inBG.name
    if '.' in source_name:
        split = source_name.split(".", 1)
        source_name = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H-%M-%S')
    
    FixWH = 4000-int(500*inEvidence)
    
    Ratio = FixWH*1.00/OrigWH*1.00
    rWidth = int(width*Ratio)
    rHeight = int(height*Ratio)
    
    NewImage = pdb.gimp_image_new (width, height, RGB)
    isamask = pdb.gimp_item_is_layer_mask (inBG)
    if isamask==True:    
        inLayer = pdb.gimp_layer_from_mask (inBG)
        NewBG = pdb.gimp_layer_new_from_drawable (inLayer, NewImage)
    else:    
        NewBG = pdb.gimp_layer_new_from_drawable (inBG, NewImage)
    NewImage.add_layer (NewBG, 0)
    pdb.gimp_layer_add_alpha (NewBG)    # RGBA
    NewOrig = NewBG.copy()
    NewImage.add_layer (NewOrig, 1)
    NewOrig.name = source_name+" (original)" 
    pdb.gimp_image_scale (NewImage, rWidth, rHeight)
    # check mean luma
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (NewBG,
                                  HISTOGRAM_LUMINANCE, 0.0, 1.0)
    #info_msg("mean luminance = "+str(meanl))
    mofified=False
    if meanl<50:
        pdb.gimp_drawable_curves_spline (NewBG, HISTOGRAM_VALUE, 6,
                                         (0.00,0.00, 0.35,0.65, 1.00,1.00)
                                         )
        mofified=True
    elif meanl<100:
        pdb.gimp_drawable_curves_spline (NewBG, HISTOGRAM_VALUE, 6,
                                         (0.00,0.00, 0.45,0.65, 1.00,1.00)
                                         )
        mofified=True
    elif meanl>150:
        pdb.gimp_drawable_curves_spline (NewBG, HISTOGRAM_VALUE, 6,
                                         (0.00,0.00, 0.65,0.45, 1.00,1.00)
                                         )
        mofified=True
    elif meanl>200:
        pdb.gimp_drawable_curves_spline (NewBG, HISTOGRAM_VALUE, 6,
                                         (0.00,0.00, 0.65,0.35, 1.00,1.00)
                                         )
        mofified=True
    if mofified==True:
        (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (NewBG,
                                      HISTOGRAM_LUMINANCE, 0.0, 1.0)
        #info_msg("modified mean luminance = "+str(meanl))
        
    Base  = StretchRGBcolours (NewBG, 0.01)
    
    Base.mode = LAYER_MODE_NORMAL
    Base.name = "Base: Image Layer Base for Cartoon Effect"
    Base.opacity = 100.0
    
    Cartooned = Base.copy()
    NewImage.add_layer (Cartooned, 0)
    Cartooned.mode = LAYER_MODE_GRAIN_MERGE       
    Cartooned.name = "Cartooned: Image Layer for Cartoon Effect"
    Cartooned.opacity = 75.0
    pdb.gimp_layer_add_alpha (Cartooned)    # RGBA
    # filter parameters fixed
    RADIUS=3.5; BLACK=0.2*inEvidence    # from 0.2 to 1.0
    pdb.plug_in_cartoon (NewImage, Cartooned, RADIUS, BLACK)
    
    Sketched = Base.copy()
    NewImage.add_layer (Sketched, 0)
    Sketched.name = "Sketched: Image Layer for Sketch Effect"
    Sketched.opacity = 50.0+(inEvidence*10.0)
    
    #pdb.gimp_display_new (NewImage)
    #pdb.gimp_context_pop()
    #return
    
    #[G'MIC] Multi-Layer Etch: fx_gcd_layeretch 10,1,16,0.12,25,7,5,0,0,3,2,1,0  last  
    Preset="10,1,16,0.12,25,7,5,0,0,3,2,1,0 "
    pdb.plug_in_gmic_qt(NewImage, Sketched, 1,0,
        "-v - -fx_gcd_layeretch "+Preset)
    Sketched.name = "Etched: Image Layer for Etch Effect "+Preset
    Sketched.mode = LAYER_MODE_MULTIPLY #GRAIN_MERGE
    
    # ================================================================
    # combine to get the drawing on a sheet of paper
    # ----------------------------------------------------------------
    # colour selection
    if predefcol == 0:
        (uR,uG,uB,A)=inColor
        pdb.gimp_context_set_foreground (inColor);
        po=" paper specified by user ("+str(int(uR))+","+str(int(uG))+","+str(int(uB))+")"
    if predefcol == 1:
        pdb.gimp_context_set_foreground (NEUTRAL_GRAY); po=" paper NEUTRAL_GRAY"
    if predefcol == 2:
        pdb.gimp_context_set_foreground (RED_GRAY); po=" paper RED_GRAY"
    if predefcol == 3:
        pdb.gimp_context_set_foreground (ORANGE_GRAY); po=" paper ORANGE_GRAY"
    if predefcol == 4:
        pdb.gimp_context_set_foreground (YELLOW_GRAY); po=" paper YELLOW_GRAY"
    if predefcol == 5:
        pdb.gimp_context_set_foreground (CHARTREUSE_GRAY); po=" paper CHARTREUSE_GRAY"
    if predefcol == 6:
        pdb.gimp_context_set_foreground (GREEN_GRAY); po=" paper GREEN_GRAY"
    if predefcol == 7:
        pdb.gimp_context_set_foreground (SPRING_GRAY); po=" paper SPRING_GRAY"
    if predefcol == 8:
        pdb.gimp_context_set_foreground (CYAN_GRAY); po=" paper CYAN_GRAY"
    if predefcol == 9:
        pdb.gimp_context_set_foreground (AZURE_GRAY); po=" paper AZURE_GRAY"
    if predefcol == 10:
        pdb.gimp_context_set_foreground (BLUE_GRAY); po=" paper BLUE_GRAY"
    if predefcol == 11:
        pdb.gimp_context_set_foreground (VIOLET_GRAY); po=" paper VIOLET_GRAY"
    if predefcol == 12:
        pdb.gimp_context_set_foreground (MAGENTA_GRAY); po=" paper MAGENTA_GRAY"
    if predefcol == 13:
        pdb.gimp_context_set_foreground (ROSE_GRAY); po=" paper ROSE_GRAY"
        
    pdb.gimp_image_select_color (NewImage, CHANNEL_OP_REPLACE,
                                 Cartooned, (0.0,0.0,0.0,1.0))
    pdb.gimp_selection_invert (NewImage)
    pdb.gimp_drawable_edit_fill (Cartooned, FILL_FOREGROUND)
    pdb.gimp_selection_none (NewImage)
    
    pdb.gimp_image_select_color (NewImage, CHANNEL_OP_REPLACE,
                                 Sketched, (1.0,1.0,1.0,1.0))
    pdb.gimp_drawable_edit_fill (Sketched, FILL_FOREGROUND)
    pdb.gimp_selection_none (NewImage)
    
    # obtain HSV from Colour RGB to colorize paper
    (R,G,B,A)=pdb.gimp_context_get_foreground()
    hue,sat,val = rgb2hsv ((R,G,B,A))
    # create the paper with the user input colour
    Paper = CreatePaper (rWidth, rHeight, hue, sat, val)
    PaperBG = pdb.gimp_layer_new_from_drawable (Paper, NewImage)
    NewImage.add_layer (PaperBG, 2)
    PaperBG.name = "Generated Paper"
    PaperBG.mode = LAYER_MODE_LINEAR_LIGHT
    PaperBG.opacity = 90.0
    # at end desaturate the Base (better look as drawing)
    pdb.gimp_drawable_desaturate (Base, DESATURATE_LUMA)
    Base.opacity = 100.0
     
    BGonTop = NewOrig.copy()
    NewImage.add_layer (BGonTop, 0)
    BGonTop.name = source_name+" (original for colours overlay)"
    
    BGonTop.mode = LAYER_MODE_LCH_COLOR
    BGonTop.opacity = inOpacity*1.0 # float
    
    # ================================================================
    # outcome 
    # ----------------------------------------------------------------
    
    if BGonTop.opacity==0: look=", grayscale"
    elif BGonTop.opacity==100: look=", coloured"
    else: look=", original at "+str(inOpacity)+"%"
    
    strokes= ", strokes evidence="+str(int(inEvidence)) 
    
    suffix = ", drawed in style E, on "+po+look+strokes
    Outcome = pdb.gimp_layer_new_from_visible (NewImage, NewImage, "")
    NewImage.add_layer (Outcome, 0)
    Outcome.name = st+" "+source_name+suffix
    pdb.gimp_item_set_visible (Outcome, False)
    
    if outimgWH==0: outimgWH=OrigWH
    uRatio = outimgWH*1.00/OrigWH*1.00
    uWidth = int(width*uRatio)
    uHeight = int(height*uRatio)
    
    pdb.gimp_image_scale (NewImage, uWidth, uHeight)   # to user chosen size
        
    pdb.gimp_display_new (NewImage)
    
    pdb.gimp_context_pop()
     
    return(Outcome)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="dinasset"
menu="<Image>/Diego/Drawing/"
copyrightYear="2020"
desc="Draw Effect Style E"

register(
    "ExperimentingE",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    "Draw Effect Style E...",
    "*",
    [
      (PF_IMAGE, "image", "Input image", None),
      (PF_DRAWABLE, "idrawable", "Input drawable", None),
      (PF_ADJUSTMENT, "outwh", "Output image Width+Heigth"+
                        "\n (0=same as input)", 4000, (0, 10000, 100)),
      (PF_OPTION, "precol", "Predefined paper colour", 0, ["*user (choose below)",
                                                        "NEUTRAL_GRAY=(170,170,170,255)",
                                                        "RED_GRAY=(200,170,170,255)",
                                                        "ORANGE_GRAY=(200,185,170,255)",
                                                        "YELLOW_GRAY=(200,200,170,255)",
                                                        "CHARTREUSE_GRAY=(185,200,170,255)",
                                                        "GREEN_GRAY=(170,200,170,255)",
                                                        "SPRING_GRAY=(170,200,185,255)",
                                                        "CYAN_GRAY=(170,200,200,255)",
                                                        "AZURE_GRAY=(170,185,200,255)",
                                                        "BLUE_GRAY=(170,170,200,255)",
                                                        "VIOLET_GRAY=(185,170,200,255)",
                                                        "MAGENTA_GRAY=(200,170,200,255)",
                                                        "ROSE_GRAY=(200,170,185,255)"
                                                        ]),
      (PF_COLOR, "pcolor", "User paper color"+
                           "\n (use light colours)", (160,160,160,255)),
      (PF_ADJUSTMENT, "strokevi", "Evidence of strokes"+
                      "\n (1=minimum, 2-4=linear increase. 5=maximm)", 3, (1, 5, 1)),
      (PF_ADJUSTMENT, "bgontop", "BG on top opacity"+
                      "\n (0=no influence, 1-100=linear influence)", 0, (0, 100, 1)),
    ],
    [
      (PF_DRAWABLE, "odrawable", "Output drawable", None),  
    ],
    DrawStyleE,
    menu=menu
    )

main()
